﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Windows.Forms;
using Lib3D;

namespace Extrude
{
    public class Scene2D : Scene
    {
        private readonly double GridSize = 50;
        private readonly int GridDim = 101;

        private SceneNode _Polygons;
        private SceneNode _BgPolygons;
        private double _Zoom;
        private Polygon _SelPolygon;

        public Scene2D(Scene2D scene)
            : base(scene)
        {
            _BgPolygons = Last.Previous;
            _Polygons = Last;
            _Zoom = 500;
            _SelPolygon = null;
        }

        public Scene2D(Control viewControl)
            : base(viewControl)
        {
            Initialize();
        }

        public override void Clear()
        {
            _Polygons.ClearChildren();
            _BgPolygons.ClearChildren();
            _SelPolygon = null;

            InitializeCamera();
        }

        public void SwapWinding()
        {
            if (_SelPolygon != null)
                _SelPolygon.SwapWinding();
        }

        public void Load(string file, bool bg)
        {
            string[] lines = File.ReadAllLines(file);
            Polygon polygon = null;

            if (bg)
                _BgPolygons.ClearChildren();
            else
                Clear();
            foreach (string line in lines)
            {
                string l = line.Trim();

                if (l.Length == 0 || l.StartsWith("#"))
                    continue;
                if (l == "p")
                {
                    if (polygon != null)
                        polygon.CreateGeometryBuffers(!bg);
                    polygon = new Polygon(GraphicsSystem);
                    if (bg)
                        _BgPolygons.AddChild(polygon);
                    else
                        _Polygons.AddChild(polygon);
                }
                else
                    polygon.Points.Add(new Vector2(l));
            }
            if (polygon != null)
                polygon.CreateGeometryBuffers(!bg);
            _SelPolygon = null;
        }

        public void Save(string file)
        {
            int count = 0;

            foreach (Polygon polygon in _Polygons)
                count += polygon.Points.Count + 1;

            List<string> lines = new List<string>(count + 1);

            lines.Add("# Polygon dump from Extrude program by McNamara Consulting, Inc. @2010");
            foreach (Polygon polygon in _Polygons)
            {
                lines.Add("p");
                foreach (Vector2 pt in polygon.Points)
                    lines.Add(string.Format("{0} {1}", pt.X, pt.Y));
            }
            File.WriteAllLines(file, lines.ToArray());
        }

        public void Extrude(Scene3D scene, double height, double resolution)
        {
            scene.Extrude(_Polygons, height, resolution);
        }

        public override void Draw()
        {
            GraphicsSystem.Draw();
        }

        public void MouseDown(MouseEventArgs e, bool snap, bool add, bool delete)
        {
            Matrix m = Camera.ProjectionMatrix * Camera.ViewMatrix;
            Vector3 p = m.Unproject(e.X, ViewControl.ClientSize.Height - e.Y, 0, ViewControl.ClientSize.Width, ViewControl.ClientSize.Height);
            double x = p.X;
            double y = p.Z;

            if (snap)
            {
                x = Math.Round(x / GridSize) * GridSize;
                y = Math.Round(y / GridSize) * GridSize;
            }

            Button = e.Button;
            MouseLast = new Vector2(e.X, e.Y);

            if (_SelPolygon != null)
                _SelPolygon.Sel = -1;

            if (add)
            {
                if (_SelPolygon != null)
                    _SelPolygon.Points.Add(new Vector2(x, y));
                else
                {
                    _Polygons.AddChild(_SelPolygon = new Polygon(GraphicsSystem));
                    _SelPolygon.Points.Add(new Vector2(x, y));
                }
                _SelPolygon.CreateGeometryBuffers(true);
            }
            else
            {
                _SelPolygon = null;

                foreach (Polygon polygon in _Polygons)
                {
                    polygon.Sel = -1;

                    for (int j = 0; j != polygon.Points.Count; j++)
                    {
                        double pX = polygon.Points[j].X;
                        double pY = polygon.Points[j].Y;

                        if (p.X >= pX - 10 && p.X <= pX + 10 && p.Z >= pY - 10 && p.Z <= pY + 10)
                        {
                            polygon.Sel = j;
                            break;
                        }
                    }
                    if (polygon.Sel != -1)
                    {
                        if (delete)
                        {
                            polygon.Points.RemoveAt(polygon.Sel);
                            polygon.Sel = -1;
                            if (polygon.Points.Count == 0)
                                polygon.Detach();
                            polygon.CreateGeometryBuffers(true);
                        }
                        else
                            _SelPolygon = polygon;
                        break;
                    }
                }
            }
        }

        public void MouseUp(MouseEventArgs e)
        {
            Button = MouseButtons.None;
            if (_SelPolygon != null)
                _SelPolygon.Sel = -1;
        }

        public void MouseMove(MouseEventArgs e, bool snap, bool camera, bool move)
        {
            Vector2 p = new Vector2(e.X, e.Y);
            Vector2 d = p - MouseLast;

            MouseLast = p;
            if (camera)
            {
                if (Button == MouseButtons.Left)
                {
                    Camera.Position += new Vector3(d.X, 0, d.Y) * 5;
                    Camera.Target += new Vector3(d.X, 0, d.Y) * 5;
                }
                else if (Button == MouseButtons.Right)
                    _Zoom += d.Y * 10;
                SetProjection();
            }
            else if (move && _SelPolygon != null)
            {
                Polygon polygon = _SelPolygon;

                if (polygon.Sel != -1)
                {
                    Matrix m = Camera.ProjectionMatrix * Camera.ViewMatrix;
                    Vector3 p2 = m.Unproject(e.X, ViewControl.ClientSize.Height - e.Y, 0, ViewControl.ClientSize.Width, ViewControl.ClientSize.Height);
                    double x = p2.X;
                    double y = p2.Z;

                    if (snap)
                    {
                        x = Math.Round(x / GridSize) * GridSize;
                        y = Math.Round(y / GridSize) * GridSize;
                    }

                    polygon.Points[polygon.Sel] = new Vector2(x, y);
                    polygon.CreateGeometryBuffers(true);
                }
            }
        }

        protected override void InitializeCamera()
        {
            GraphicsSystem.Camera = new Camera();
            Camera.OrthoProjection = true;
            Camera.Position = new Vector3(0, 1, 0);
            Camera.Target = new Vector3(0, 0, 0);
            Camera.Up = new Vector3(0, 0, 1);
            Camera.NearPlane = -100;
            Camera.FarPlane = 100;

            _Zoom = 500;

            SetProjection();
        }

        private void SetProjection()
        {
            double aspect = GraphicsSystem.Aspect;

            Camera.Left = -_Zoom * aspect;
            Camera.Right = _Zoom * aspect;
            Camera.Bottom = -_Zoom;
            Camera.Top = _Zoom;
        }

        private void Initialize()
        {
            double a = -GridDim / 2 * GridSize;
            double b = GridDim / 2 * GridSize;
            GeometryBuffer geometryBuffer = new GeometryBuffer(GraphicsSystem);

            geometryBuffer.PrimitiveType = PrimitiveType.Lines;
            geometryBuffer.Vertices = new List<Vertex>(GridDim * 4);
            geometryBuffer.Indices = new List<int>(GridDim * 4);
            for (int i = 0, j = 0; i != GridDim; i++)
            {
                double c = a + i * GridSize;

                geometryBuffer.Vertices.Add(new Vertex(new Vector3(c, 0, a), new Vector3(), new Vector2()));
                geometryBuffer.Vertices.Add(new Vertex(new Vector3(c, 0, b), new Vector3(), new Vector2()));
                geometryBuffer.Vertices.Add(new Vertex(new Vector3(a, 0, c), new Vector3(), new Vector2()));
                geometryBuffer.Vertices.Add(new Vertex(new Vector3(b, 0, c), new Vector3(), new Vector2()));
                for (int k = 0; k != 4; k++, j++)
                    geometryBuffer.Indices.Add(j);

            }
            Materials[0].LineWidth = 1;
            Materials[0].LightingEnabled = false;
            Materials[0].BlendEnabled = false;
            Materials[0].DiffuseColor = new Vector4(0.5, 0.5, 0.5, 1);
            Materials[0].GeometryBuffer = geometryBuffer;
            Materials[0].Pass = RenderPass.Immediate;

            SceneNode axisNode = new SceneNode(GraphicsSystem);

            geometryBuffer = new GeometryBuffer(GraphicsSystem);
            geometryBuffer.PrimitiveType = PrimitiveType.Lines;

            geometryBuffer.Vertices = new List<Vertex>(
                new Vertex[]
                {
                    new Vertex(new Vector3(0, 0, a), new Vector3(), new Vector2()),
                    new Vertex(new Vector3(0, 0, b), new Vector3(), new Vector2()),
                    new Vertex(new Vector3(a, 0, 0), new Vector3(), new Vector2()),
                    new Vertex(new Vector3(b, 0, 0), new Vector3(), new Vector2()),
                }
                );
            geometryBuffer.Indices = new List<int>(new int[] { 0, 1, 2, 3 });
            axisNode.Materials[0].LightingEnabled = false;
            axisNode.Materials[0].BlendEnabled = false;
            axisNode.Materials[0].LineWidth = 2;
            axisNode.Materials[0].DiffuseColor = new Vector4(1, 0.5, 0, 1);
            axisNode.Materials[0].GeometryBuffer = geometryBuffer;
            axisNode.Materials[0].Pass = RenderPass.Immediate;
            AddChild(axisNode);

            AddChild(_BgPolygons = new SceneNode(GraphicsSystem));
            AddChild(_Polygons = new SceneNode(GraphicsSystem));

            _SelPolygon = null;
        }
    }
}
