﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Windows.Forms;
using Lib3D;

namespace Extrude
{
    public class Scene3D : Scene
    {
        private SceneNode _Mesh;
        private TextureLoader _TexLoader;

        public Scene3D(Scene3D scene)
            : base(scene)
        {
            _Mesh = First;
            _TexLoader = scene._TexLoader;
        }

        public Scene3D(Control viewControl)
            : base(viewControl)
        {

            _Mesh = new SceneNode(GraphicsSystem);
            _Mesh.Materials[0].Name = "mat1";
            _TexLoader = new TextureLoader(GraphicsSystem);

            AddChild(_Mesh);

            AddChild(new LightSceneNode(GraphicsSystem, new Vector3(800, 800, 800), new Vector4(1, 1, 1, 1), 0.0001f));
            AddChild(new LightSceneNode(GraphicsSystem, new Vector3(-800, -800, -800), new Vector4(1, 1, 1, 1), 0.0001f));
        }

        public void SetTexture(string file)
        {
            _Mesh.Materials[0].Texture = null;
            _TexLoader.Clear();
            if (File.Exists(file))
                _Mesh.Materials[0].Texture = _TexLoader.Get(file);
        }

        public override void Clear()
        {
            _Mesh.Materials[0].GeometryBuffer.Dispose();
            _Mesh.Materials[0].GeometryBuffer = new GeometryBuffer(GraphicsSystem);
            _Mesh.Materials[0].Texture = null;
            _TexLoader.Clear();
            InitializeCamera();
        }

        public override void Draw()
        {
            GraphicsSystem.Draw();
        }

        public void MouseDown(MouseEventArgs e)
        {
            Button = e.Button;
            MouseLast = new Vector2(e.X, e.Y);
        }

        public void MouseUp(MouseEventArgs e)
        {
            Button = MouseButtons.None;
        }

        public void MouseMove(MouseEventArgs e)
        {
            if (Button == MouseButtons.Left || Button == MouseButtons.Right)
            {
                Vector2 p = new Vector2(e.X, e.Y);
                Vector2 d = p - MouseLast;

                MouseLast = p;

                if (Button == MouseButtons.Left)
                    Camera.RotateAroundTarget(d);
                else
                {
                    Vector3 f = Camera.Position - Camera.Target;

                    f = Vector3.Normalize(f) * (f.Length() + Math.Sign(d.Y) * 20);
                    Camera.Position = Camera.Target + f;
                }
            }
        }

        protected override void InitializeCamera()
        {
            GraphicsSystem.Camera = new Camera();
            Camera.Position = new Vector3(500, 500, 500);
            Camera.Target = new Vector3(0, 0, 0);
            Camera.Up = new Vector3(-1, 1, -1);
            Camera.FOV = 60;
            Camera.Aspect = GraphicsSystem.Aspect;
            Camera.NearPlane = 1;
            Camera.FarPlane = 5000;
        }

        public void Save(string file)
        {
            OBJFile.Write(file, _Mesh.Materials);
        }

        public void Extrude(SceneNode polygons, double height, double resolution)
        {
            List<GeometryBuffer> geometryBuffers = new List<GeometryBuffer>();
            GeometryBuffer geometryBuffer;

            _Mesh.Materials[0].GeometryBuffer.Dispose();

            foreach (Polygon polygon in polygons)
            {
                List<Vector2> points = polygon.Points;

                if (polygon.GetArea() > 0)
                    polygon.SwapWinding(points = new List<Vector2>(points));
                geometryBuffer = new GeometryBuffer(GraphicsSystem);
                geometryBuffer.Extrude(points, height);
                geometryBuffer.CalcTextureCoordinates(Matrix.CreateIdentity(), resolution);
                geometryBuffers.Add(geometryBuffer);
            }
            geometryBuffer = new GeometryBuffer(GraphicsSystem);
            for (int i = 0; i != geometryBuffers.Count; i++)
            {
                int x = geometryBuffer.Vertices.Count;

                geometryBuffer.Vertices.AddRange(geometryBuffers[i].Vertices);
                for (int j = 0; j != geometryBuffers[i].Indices.Count; j++)
                    geometryBuffer.Indices.Add(geometryBuffers[i].Indices[j] + x);
            }
            geometryBuffer.Compile();

            _Mesh.Materials[0].GeometryBuffer = geometryBuffer;
        }

        public override void Dispose()
        {
            _Mesh.Materials[0].Texture = null;
            _Mesh.Materials[0].GeometryBuffer.Dispose();
            _TexLoader.Clear();

            base.Dispose();
        }
    }
}
