﻿using System;

namespace Lib3D
{
    public struct AABBox
    {
        public Vector3 Min;
        public Vector3 Max;

        public AABBox(Vector3 min, Vector3 max)
        {
            Min = min;
            Max = max;
        }

        public bool IsEmpty
        {
            get { return Min.X > Max.X && Min.Y > Max.Y && Min.Z > Max.Z; }
        }

        public Vector3 Dimension
        {
            get { return Max - Min; }
        }

        public Vector3 Center
        {
            get { return (Max + Min) / 2; }
        }

        public void Clear()
        {
            double d = double.MaxValue;

            Min = new Vector3(d, d, d);
            Max = -Min;
        }

        public Nullable<double> ISect(Ray ray)
        {
            double tNear = -double.MaxValue;
            double tFar = double.MaxValue;
            double t1, t2, temp;

            if (ray.Direction.X == 0 && (ray.Origin.X < Min.X || ray.Origin.X > Max.X))
                return null;
            else if (ray.Direction.X != 0)
            {
                t1 = (Min.X - ray.Origin.X) / ray.Direction.X;
                t2 = (Max.X - ray.Origin.X) / ray.Direction.X;
                if (t1 > t2)
                {
                    temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return null;
                if (tFar < 0)
                    return null;
            }

            if (ray.Direction.Y == 0 && (ray.Origin.Y < Min.Y || ray.Origin.Y > Max.Y))
                return null;
            else if (ray.Direction.Y != 0)
            {
                t1 = (Min.Y - ray.Origin.Y) / ray.Direction.Y;
                t2 = (Max.Y - ray.Origin.Y) / ray.Direction.Y;
                if (t1 > t2)
                {
                    temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return null;
                if (tFar < 0)
                    return null;
            }

            if (ray.Direction.Z == 0 && (ray.Origin.Z < Min.Z || ray.Origin.Z > Max.Z))
                return null;
            else if (ray.Direction.Z != 0)
            {
                t1 = (Min.Z - ray.Origin.Z) / ray.Direction.Z;
                t2 = (Max.Z - ray.Origin.Z) / ray.Direction.Z;
                if (t1 > t2)
                {
                    temp = t1;
                    t1 = t2;
                    t2 = temp;
                }
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return null;
                if (tFar < 0)
                    return null;
            }
            return tNear;
        }

        public static AABBox CreateEmpty()
        {
            AABBox box = new AABBox();

            box.Clear();

            return box;
        }

        public static AABBox operator +(AABBox a, Vector3 b)
        {
            if (b.X < a.Min.X)
                a.Min.X = b.X;
            if (b.X > a.Max.X)
                a.Max.X = b.X;

            if (b.Y < a.Min.Y)
                a.Min.Y = b.Y;
            if (b.Y > a.Max.Y)
                a.Max.Y = b.Y;

            if (b.Z < a.Min.Z)
                a.Min.Z = b.Z;
            if (b.Z > a.Max.Z)
                a.Max.Z = b.Z;

            return a;
        }

        public static AABBox operator +(AABBox a, AABBox b)
        {
            if (!b.IsEmpty)
            {
                if (b.Min.X < a.Min.X)
                    a.Min.X = b.Min.X;
                if (b.Max.X > a.Max.X)
                    a.Max.X = b.Max.X;

                if (b.Min.Y < a.Min.Y)
                    a.Min.Y = b.Min.Y;
                if (b.Max.Y > a.Max.Y)
                    a.Max.Y = b.Max.Y;

                if (b.Min.Z < a.Min.Z)
                    a.Min.Z = b.Min.Z;
                if (b.Max.Z > a.Max.Z)
                    a.Max.Z = b.Max.Z;
            }
            return a;
        }

        public static AABBox operator *(Matrix a, AABBox b)
        {
            if (b.IsEmpty)
                return b;

            Vector3 translation = a.Translation;
            AABBox t = new AABBox(translation, translation);

            t.Min.X += (a.M00 < 0) ? a.M00 * b.Max.X : a.M00 * b.Min.X;
            t.Min.X += (a.M01 < 0) ? a.M01 * b.Max.Y : a.M01 * b.Min.Y;
            t.Min.X += (a.M02 < 0) ? a.M02 * b.Max.Z : a.M02 * b.Min.Z;

            t.Max.X += (a.M00 > 0) ? a.M00 * b.Max.X : a.M00 * b.Min.X;
            t.Max.X += (a.M01 > 0) ? a.M01 * b.Max.Y : a.M01 * b.Min.Y;
            t.Max.X += (a.M02 > 0) ? a.M02 * b.Max.Z : a.M02 * b.Min.Z;

            t.Min.Y += (a.M10 < 0) ? a.M10 * b.Max.X : a.M10 * b.Min.X;
            t.Min.Y += (a.M11 < 0) ? a.M11 * b.Max.Y : a.M11 * b.Min.Y;
            t.Min.Y += (a.M12 < 0) ? a.M12 * b.Max.Z : a.M12 * b.Min.Z;

            t.Max.Y += (a.M10 > 0) ? a.M10 * b.Max.X : a.M10 * b.Min.X;
            t.Max.Y += (a.M11 > 0) ? a.M11 * b.Max.Y : a.M11 * b.Min.Y;
            t.Max.Y += (a.M12 > 0) ? a.M12 * b.Max.Z : a.M12 * b.Min.Z;

            t.Min.Z += (a.M20 < 0) ? a.M20 * b.Max.X : a.M20 * b.Min.X;
            t.Min.Z += (a.M21 < 0) ? a.M21 * b.Max.Y : a.M21 * b.Min.Y;
            t.Min.Z += (a.M22 < 0) ? a.M22 * b.Max.Z : a.M22 * b.Min.Z;

            t.Max.Z += (a.M20 > 0) ? a.M20 * b.Max.X : a.M20 * b.Min.X;
            t.Max.Z += (a.M21 > 0) ? a.M21 * b.Max.Y : a.M21 * b.Min.Y;
            t.Max.Z += (a.M22 > 0) ? a.M22 * b.Max.Z : a.M22 * b.Min.Z;

            return t;
        }
    }
}
