﻿using System;
using System.Collections.Generic;

namespace Lib3D
{
    public enum PrimitiveType
    {
        Points = 0,
        Lines = 0x0001,
        Triangles = 0x0004,
        Quads = 0x0007,
        Polygon = 0x0009
    }

    public class GeometryBuffer : IDisposable
    {
        private GraphicsSystem _GraphicsSystem;
        private uint _List;

        public GeometryBuffer(GraphicsSystem graphicsSystem)
        {
            _GraphicsSystem = graphicsSystem;
            _List = 0;

            PrimitiveType = PrimitiveType.Triangles;
            Name = "";
            VertexColoringEnabled = false;
            Vertices = new List<Vertex>();
            Indices = new List<int>();
            Bounds = AABBox.CreateEmpty();
        }

        public GeometryBuffer(GeometryBuffer geometryBuffer)
        {
            _GraphicsSystem = geometryBuffer._GraphicsSystem;
            _List = 0;

            PrimitiveType = geometryBuffer.PrimitiveType;
            Name = geometryBuffer.Name;
            VertexColoringEnabled = geometryBuffer.VertexColoringEnabled;
            Vertices = new List<Vertex>(geometryBuffer.Vertices);
            Indices = new List<int>(geometryBuffer.Indices);
            Bounds = geometryBuffer.Bounds;
        }

        public GraphicsSystem GraphicsSystem
        {
            get { return _GraphicsSystem; }
        }

        public PrimitiveType PrimitiveType
        {
            get;
            set;
        }

        public string Name
        {
            get;
            set;
        }

        public bool VertexColoringEnabled
        {
            get;
            set;
        }

        public List<Vertex> Vertices
        {
            get;
            set;
        }

        public List<int> Indices
        {
            get;
            set;
        }

        public AABBox Bounds
        {
            get;
            set;
        }

        public void CalcBounds()
        {
            Bounds = AABBox.CreateEmpty();
            for (int i = 0; i != Vertices.Count; i++)
                Bounds += Vertices[i].Position;
        }

        public void Explode()
        {
            List<Vertex> vertices = new List<Vertex>(Indices.Count);

            for (int i = 0; i != Indices.Count; i++)
                vertices.Add(Vertices[Indices[i]]);
            for (int i = 0; i != Indices.Count; i++)
                Indices[i] = i;
            Vertices = vertices;
        }

        public void CalcTextureCoordinates(Matrix matrix, double resolution)
        {
            Matrix matrixIT = Matrix.Transpose(Matrix.Invert(matrix));

            for (int i = 0; i != Vertices.Count; i++)
            {
                Vertex vertex = Vertices[i];
                Vector3 n = Vector3.Normalize(matrixIT * vertex.Normal);
                double x = Math.Abs(n.X);
                double y = Math.Abs(n.Y);
                double z = Math.Abs(n.Z);
                Vector3 p = matrix * vertex.Position;
                Vector2 t;

                if (x > y && x > z)
                    t = new Vector2(p.Z, p.Y) * resolution;
                else if (y > x && y > z)
                    t = new Vector2(p.X, p.Z) * resolution;
                else
                    t = new Vector2(p.X, p.Y) * resolution;
                vertex.TextureCoordinate = t;

                Vertices[i] = vertex;
            }
        }

        public void Transform(Matrix matrix)
        {
            Matrix matrixIT = Matrix.Transpose(Matrix.Invert(matrix));

            for (int i = 0; i != Vertices.Count; i++)
            {
                Vertex vertex = Vertices[i];

                vertex.Position = matrix * vertex.Position;
                vertex.Normal = matrixIT * vertex.Normal;
                Vertices[i] = vertex;
            }
        }

        public void CalcNormals()
        {
            for (int i = 0; i != Vertices.Count; i++)
            {
                Vertex vertex = Vertices[i];

                vertex.Normal = new Vector3(0, 0, 0);
                Vertices[i] = vertex;
            }

            for (int i = 0; i != Indices.Count; )
            {
                Vector3 a = Vertices[Indices[i]].Position;
                Vector3 b = Vertices[Indices[i + 1]].Position;
                Vector3 c = Vertices[Indices[i + 2]].Position;
                Vector3 n = Vector3.Normalize(Vector3.Cross(c - b, b - a));

                for (int j = 0; j != 3; j++, i++)
                {
                    Vertex vertex = Vertices[Indices[i]];

                    vertex.Normal += n;
                    Vertices[Indices[i]] = vertex;
                }
            }

            for (int i = 0; i != Vertices.Count; i++)
            {
                Vertex vertex = Vertices[i];

                vertex.Normal.Normalize();
                Vertices[i] = vertex;
            }
        }

        public void SwapWinding()
        {
            int i = 0;
            int j = Indices.Count - 1;

            while (i < j)
            {
                int temp = Indices[i];

                Indices[i] = Indices[j];
                Indices[j] = temp;

                i++;
                j--;
            }
        }

        public void FlipNormals()
        {
            for (int i = 0; i != Vertices.Count; i++)
            {
                Vertex vertex = Vertices[i];

                vertex.Normal = -vertex.Normal;
                Vertices[i] = vertex;
            }
        }

        public void Triangulate(List<Vector2> polygon, double y)
        {
            Triangulator triangulator = new Triangulator(polygon);

            PrimitiveType = PrimitiveType.Triangles;
            Vertices.Clear();
            Indices.Clear();
            Vertices.Capacity = Indices.Capacity = triangulator.Triangles.Count;

            for (int i = 0; i != triangulator.Triangles.Count; i++)
            {
                Vector2 pt = triangulator.Triangles[i];

                Vertices.Add(new Vertex(new Vector3(pt.X, y, pt.Y), new Vector3(0, 1, 0), new Vector2(0, 0)));
                Indices.Add(i);
            }
        }

        public void Extrude(List<Vector2> polygon, double height)
        {
            double d = height / 2;
            int idx = 0;

            PrimitiveType = PrimitiveType.Triangles;

            Vertices.Clear();
            Indices.Clear();

            for (int i = 0; i != polygon.Count; i++)
            {
                Vector2 b = polygon[i];
                Vector2 a = polygon[(i + 1) % polygon.Count];
                Vector2 n = b - a;

                n.Normalize();

                Vector3 n2 = new Vector3(-n.Y, 0, n.X);

                Vertices.Add(new Vertex(new Vector3(a.X, -d, a.Y), n2, new Vector2()));
                Vertices.Add(new Vertex(new Vector3(a.X, d, a.Y), n2, new Vector2()));
                Vertices.Add(new Vertex(new Vector3(b.X, d, b.Y), n2, new Vector2()));
                Vertices.Add(new Vertex(new Vector3(b.X, -d, b.Y), n2, new Vector2()));


                Indices.Add(idx);
                Indices.Add(idx + 1);
                Indices.Add(idx + 2);
                Indices.Add(idx + 2);
                Indices.Add(idx + 3);
                Indices.Add(idx);

                idx += 4;
            }

            Triangulator triangulator = new Triangulator(polygon);

            for (int i = 0; i != triangulator.Triangles.Count; )
            {
                Vector2 a = triangulator.Triangles[i++];
                Vector2 b = triangulator.Triangles[i++];
                Vector2 c = triangulator.Triangles[i++];

                Vertices.Add(new Vertex(new Vector3(a.X, d, a.Y), new Vector3(0, 1, 0), new Vector2()));
                Vertices.Add(new Vertex(new Vector3(b.X, d, b.Y), new Vector3(0, 1, 0), new Vector2()));
                Vertices.Add(new Vertex(new Vector3(c.X, d, c.Y), new Vector3(0, 1, 0), new Vector2()));

                Vertices.Add(new Vertex(new Vector3(c.X, -d, c.Y), new Vector3(0, -1, 0), new Vector2()));
                Vertices.Add(new Vertex(new Vector3(b.X, -d, b.Y), new Vector3(0, -1, 0), new Vector2()));
                Vertices.Add(new Vertex(new Vector3(a.X, -d, a.Y), new Vector3(0, -1, 0), new Vector2()));

                for (int j = 0; j != 6; j++)
                    Indices.Add(idx++);
            }
        }

        public void Compile()
        {
            _GraphicsSystem.DisposeGeometry(ref _List);
            _GraphicsSystem.CompileGeometry(this, ref _List);
        }

        public void Dispose()
        {
            _GraphicsSystem.DisposeGeometry(ref _List);
        }

        internal uint List
        {
            get { return _List; }
        }
    }
}
