﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Windows.Forms;

namespace Lib3D
{
    public class GraphicsSystem : IDisposable, IComparer<SceneNode>
    {
        private static readonly int PFD_DOUBLEBUFFER = 0x00000001;
        private static readonly int PFD_DRAW_TO_WINDOW = 0x00000004;
        private static readonly int PFD_SUPPORT_OPENGL = 0x00000020;
        private static readonly int PFD_TYPE_RGBA = 0;
        private static readonly int GL_PROJECTION = 0x1701;
        private static readonly int GL_MODELVIEW = 0x1700;
        private static readonly int GL_TEXTURE = 0x1702;
        private static readonly int GL_DEPTH_TEST = 0x0B71;
        private static readonly int GL_CULL_FACE = 0x0B44;
        private static readonly int GL_CCW = 0x0901;
        private static readonly int GL_CW = 0x0900;
        private static readonly int GL_NORMALIZE = 0x0BA1;
        private static readonly int GL_AMBIENT = 0x1200;
        private static readonly int GL_DIFFUSE = 0x1201;
        private static readonly int GL_SPECULAR = 0x1202;
        private static readonly int GL_EMISSION = 0x1600;
        private static readonly int GL_SHININESS = 0x1601;
        private static readonly int GL_LIGHTING = 0x0B50;
        private static readonly int GL_LIGHT0 = 0x4000;
        private static readonly int GL_QUADRATIC_ATTENUATION = 0x1209;
        private static readonly int GL_LINEAR_ATTENUATION = 0x1208;
        private static readonly int GL_CONSTANT_ATTENUATION = 0x1207;
        private static readonly int GL_POSITION = 0x1203;
        private static readonly int GL_BLEND = 0x0BE2;
        private static readonly int GL_COLOR_BUFFER_BIT = 0x00004000;
        private static readonly int GL_DEPTH_BUFFER_BIT = 0x00000100;
        private static readonly int GL_FRONT_AND_BACK = 0x0408;
        private static readonly int GL_RGBA = 0x1908;
        private static readonly int GL_TEXTURE_2D = 0x0DE1;
        private static readonly int GL_UNSIGNED_BYTE = 0x1401;
        private static readonly int GL_AMBIENT_AND_DIFFUSE = 0x1602;
        private static readonly int GL_COLOR_MATERIAL = 0x0B57;
        private static readonly int GL_COMPILE = 0x1300;

        private double[] _MatrixArray;
        private float[] _VectorArray;
        private Control _ViewControl;
        private IntPtr _WindowHandle;
        private IntPtr _DeviceHandle;
        private IntPtr _RenderHandle;
        private List<SceneNode> _LightSceneNodes;
        private List<SceneNode> _RenderSceneNodes;
        private int _TexCount;
        private int _ListCount;
        private Material _BoundsMaterial;

        public GraphicsSystem(Control viewControl)
        {
            PIXELFORMATDESCRIPTOR pfd = new PIXELFORMATDESCRIPTOR();

            pfd.nSize = (short)Marshal.SizeOf(pfd);
            pfd.nVersion = 1;
            pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;
            pfd.iPixelType = (byte)PFD_TYPE_RGBA;
            pfd.cColorBits = 32;
            pfd.cDepthBits = 32;

            _ViewControl = viewControl;
            _WindowHandle = viewControl.Handle;
            _DeviceHandle = GetDC(_WindowHandle);

            int format = ChoosePixelFormat(_DeviceHandle, ref pfd);

            SetPixelFormat(_DeviceHandle, format, ref pfd);

            _RenderHandle = wglCreateContext(_DeviceHandle);
            wglMakeCurrent(_DeviceHandle, _RenderHandle);

            LoadExtensions();

            glEnable(GL_NORMALIZE);

            _MatrixArray = new double[16];
            _VectorArray = new float[4];

            _LightSceneNodes = new List<SceneNode>();
            _RenderSceneNodes = new List<SceneNode>();
            _TexCount = 0;
            _ListCount = 0;

            _BoundsMaterial = new Material(this);
            _BoundsMaterial.LightingEnabled = false;
            _BoundsMaterial.GeometryBuffer = CreateUnitBounds();

            Camera = new Camera();
            ClearColor = new Vector4(0, 0, 0, 1);
            RootNode = null;
        }

        public SceneNode RootNode
        {
            get;
            set;
        }

        public Camera Camera
        {
            get;
            set;
        }

        public Vector4 ClearColor
        {
            get;
            set;
        }

        public Control ViewControl
        {
            get { return _ViewControl; }
        }

        public double Aspect
        {
            get { return _ViewControl.ClientSize.Width / (double)_ViewControl.ClientSize.Height; }
        }

        public void SetFixedFrameRate(bool fixedRate)
        {
            if (_wglSwapIntervalEXT != null)
            {
                if (fixedRate)
                    _wglSwapIntervalEXT(1);
                else
                    _wglSwapIntervalEXT(0);
            }
        }

        public int Compare(SceneNode a, SceneNode b)
        {
            if (a == b)
                return 0;
            else
            {
                double dA = (a.AbsoluteMatrix.Translation - Camera.Position).LengthSquared();
                double dB = (b.AbsoluteMatrix.Translation - Camera.Position).LengthSquared();

                if (dA < dB)
                    return -1;
                else if (dA > dB)
                    return 1;
                else
                    return 0;
            }
        }

        public int Draw()
        {
            Matrix viewMatrix, projectionMatrix;
            Frustum frustum;

            wglMakeCurrent(_DeviceHandle, _RenderHandle);

            glViewport(0, 0, ViewControl.ClientSize.Width, ViewControl.ClientSize.Height);
            SetDepthTestEnabled(true);
            SetDepthWriteEnabled(true);
            glClearColor((float)ClearColor.X, (float)ClearColor.Y, (float)ClearColor.Z, (float)ClearColor.W);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            Camera.Aspect = Aspect;
            viewMatrix = Camera.ViewMatrix;
            projectionMatrix = Camera.ProjectionMatrix;
            frustum = new Frustum(projectionMatrix * viewMatrix);

            SetProjectionMatrix(projectionMatrix);
            SetModelViewMatrix(viewMatrix);

            _LightSceneNodes.Clear();
            _RenderSceneNodes.Clear();

            if (RootNode != null)
            {
                List<SceneNode> temp = _RenderSceneNodes;

                _LightSceneNodes.Capacity = _RenderSceneNodes.Capacity = RootNode.Count;

                RootNode.Update();
                RootNode.CalcBounds();
                AddSceneNodeToLists(RootNode, frustum, true);

                _LightSceneNodes.Sort(this);
                SetLights();

                SetDepthWriteEnabled(false);
                Draw(RenderPass.Background, viewMatrix);

                SetDepthWriteEnabled(true);
                Draw(RenderPass.Solid, viewMatrix);

                SetDepthWriteEnabled(false);
                _RenderSceneNodes = new List<SceneNode>(_RenderSceneNodes);
                _RenderSceneNodes.Sort(this);
                Draw(RenderPass.Transparent, viewMatrix);
                _RenderSceneNodes = temp;

                SetDepthTestEnabled(false);
                Draw(RenderPass.Immediate, viewMatrix);
            }
            SwapBuffers(_DeviceHandle);

            return _RenderSceneNodes.Count;
        }

        public void Dispose()
        {
            if (_RenderHandle.ToInt32() != 0)
            {
                Console.WriteLine("Disposing GraphicsSystem : Texture Count={0}, Compiled Geometry Count={1}", _TexCount, _ListCount);

                wglMakeCurrent(_DeviceHandle, IntPtr.Zero);
                wglDeleteContext(_RenderHandle);
                ReleaseDC(_WindowHandle, _DeviceHandle);

                _RenderHandle = IntPtr.Zero;
            }
        }

        internal void CreateTexture(byte[] data, int width, int height, ref uint texture)
        {
            wglMakeCurrent(_DeviceHandle, _RenderHandle);
            if (glIsTexture(texture) == 0)
            {
                texture = 0;
                glGenTextures(1, out texture);
                glBindTexture(GL_TEXTURE_2D, texture);
                gluBuild2DMipmaps(GL_TEXTURE_2D, 4, width, height, GL_RGBA, GL_UNSIGNED_BYTE, data);

                _TexCount++;
            }
        }

        internal void DisposeTexture(ref uint texture)
        {
            wglMakeCurrent(_DeviceHandle, _RenderHandle);
            if (glIsTexture(texture) != 0)
            {
                glDeleteTextures(1, ref texture);
                _TexCount--;
            }
            texture = 0;
        }

        internal void CompileGeometry(GeometryBuffer geometryBuffer, ref uint list)
        {
            wglMakeCurrent(_DeviceHandle, _RenderHandle);
            if (glIsList(list) == 0)
            {
                list = 0;
                list = glGenLists(1);
                glNewList(list, GL_COMPILE);
                Draw(geometryBuffer);
                glEndList();
                _ListCount++;
            }
        }

        internal void DisposeGeometry(ref uint list)
        {
            wglMakeCurrent(_DeviceHandle, _RenderHandle);
            if (glIsList(list) != 0)
            {
                glDeleteLists(list, 1);
                _ListCount--;
            }
            list = 0;
        }

        private void AddSceneNodeToLists(SceneNode sceneNode, Frustum frustum, bool visible)
        {
            if (visible)
            {
                visible = sceneNode.Visible && frustum.Contains(sceneNode.Bounds);
                if (visible)
                    _RenderSceneNodes.Add(sceneNode);
            }
            if (sceneNode is LightSceneNode)
                _LightSceneNodes.Add(sceneNode);
            foreach (SceneNode childNode in sceneNode)
                AddSceneNodeToLists(childNode, frustum, visible);
        }

        private void SetLights()
        {
            for (int i = 0; i != 8; i++)
                if (i < _LightSceneNodes.Count)
                {
                    SetLightEnabled(i, true);
                    SetLight(i, _LightSceneNodes[i] as LightSceneNode);
                }
                else
                    SetLightEnabled(i, false);
        }

        private void Draw(RenderPass pass, Matrix viewMatrix)
        {
            foreach (SceneNode sceneNode in _RenderSceneNodes)
            {
                Draw(sceneNode, pass, viewMatrix * sceneNode.AbsoluteMatrix);

                if (pass == RenderPass.Solid && sceneNode.DrawBounds)
                    Draw(sceneNode.Bounds, sceneNode.BoundsColor, viewMatrix);
            }
        }

        private void Draw(AABBox bounds, Vector4 color, Matrix viewMatrix)
        {
            _BoundsMaterial.DiffuseColor = color;

            SetMaterial(_BoundsMaterial);

            Matrix matrix = Matrix.CreateTranslation(bounds.Center) * Matrix.CreateScaling(bounds.Dimension);

            SetModelViewMatrix(viewMatrix * matrix);

            Draw(_BoundsMaterial.GeometryBuffer);
        }

        private void Draw(SceneNode sceneNode, RenderPass pass, Matrix modelMatrix)
        {
            SetModelViewMatrix(modelMatrix);

            foreach (Material material in sceneNode.Materials)
            {
                GeometryBuffer geometryBuffer = material.GeometryBuffer;

                if (!material.Enabled || material.Pass != pass)
                    continue;

                SetMaterial(material);

                if (glIsList(geometryBuffer.List) != 0)
                    glCallList(geometryBuffer.List);
                else
                    Draw(geometryBuffer);
            }
        }

        private void Draw(GeometryBuffer geometryBuffer)
        {
            if (geometryBuffer.VertexColoringEnabled)
            {
                glBegin((int)geometryBuffer.PrimitiveType);
                for (int i = 0; i != geometryBuffer.Indices.Count; i++)
                {
                    Vertex vertex = geometryBuffer.Vertices[geometryBuffer.Indices[i]];
                    Vector3 n = vertex.Normal;
                    Vector3 p = vertex.Position;
                    Vector4 c = vertex.Color;
                    Vector2 t = vertex.TextureCoordinate;

                    glColor4d(c.X, c.Y, c.Z, c.W);
                    glNormal3d(n.X, n.Y, n.Z);
                    glTexCoord2d(t.X, t.Y);
                    glVertex3d(p.X, p.Y, p.Z);
                }
                glEnd();
            }
            else
            {
                glBegin((int)geometryBuffer.PrimitiveType);
                for (int i = 0; i != geometryBuffer.Indices.Count; i++)
                {
                    Vertex vertex = geometryBuffer.Vertices[geometryBuffer.Indices[i]];
                    Vector3 n = vertex.Normal;
                    Vector3 p = vertex.Position;
                    Vector2 t = vertex.TextureCoordinate;

                    glNormal3d(n.X, n.Y, n.Z);
                    glTexCoord2d(t.X, t.Y);
                    glVertex3d(p.X, p.Y, p.Z);
                }
                glEnd();
            }
        }

        private void SetMaterial(Material material)
        {
            if (material.LightingEnabled)
                glEnable(GL_LIGHTING);
            else
                glDisable(GL_LIGHTING);

            if (material.ColorMaterialEnabled)
            {
                glEnable(GL_COLOR_MATERIAL);
                glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
            }
            else
                glDisable(GL_COLOR_MATERIAL);

            SetTexture(material);

            if (material.BlendEnabled)
            {
                glEnable(GL_BLEND);
                glBlendFunc((int)material.SourceBlend, (int)material.DestBlend);
            }
            else
                glDisable(GL_BLEND);

            if (material.CullMode == CullMode.None)
                glDisable(GL_CULL_FACE);
            else
            {
                glEnable(GL_CULL_FACE);
                if (material.CullMode == CullMode.Clockwise)
                    glFrontFace(GL_CCW);
                else
                    glFrontFace(GL_CW);
            }

            glLineWidth((float)material.LineWidth);
            glPointSize((float)material.PointSize);

            material.AmbientColor.CopyTo(_VectorArray);
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, _VectorArray);

            material.DiffuseColor.CopyTo(_VectorArray);
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, _VectorArray);
            glColor4d(material.DiffuseColor.X, material.DiffuseColor.Y, material.DiffuseColor.Z, material.DiffuseColor.W);

            material.SpecularColor.CopyTo(_VectorArray);
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, _VectorArray);

            material.EmissiveColor.CopyTo(_VectorArray);
            glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, _VectorArray);

            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (float)material.Power);
        }


        private void SetTexture(Material material)
        {
            if (material.TextureEnabled && material.Texture != null)
            {
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, material.Texture.ID);
                SetTextureMatrix(material.TextureMatrix);
            }
            else
                glDisable(GL_TEXTURE_2D);
        }

        private void SetDepthTestEnabled(bool enabled)
        {
            if (enabled)
                glEnable(GL_DEPTH_TEST);
            else
                glDisable(GL_DEPTH_TEST);
        }

        private void SetDepthWriteEnabled(bool enabled)
        {
            if (enabled)
                glDepthMask(1);
            else
                glDepthMask(0);
        }

        private void SetProjectionMatrix(Matrix matrix)
        {
            matrix.CopyTo(_MatrixArray);
            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            glMultMatrixd(_MatrixArray);
        }

        private void SetModelViewMatrix(Matrix matrix)
        {
            matrix.CopyTo(_MatrixArray);
            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();
            glMultMatrixd(_MatrixArray);
        }

        private void SetTextureMatrix(Matrix matrix)
        {
            matrix.CopyTo(_MatrixArray);
            glMatrixMode(GL_TEXTURE);
            glLoadIdentity();
            glMultMatrixd(_MatrixArray);
        }

        private void SetLightEnabled(int i, bool enabled)
        {
            i += GL_LIGHT0;
            if (enabled)
                glEnable(i);
            else
                glDisable(i);
        }

        private void SetLight(int i, LightSceneNode light)
        {
            i += GL_LIGHT0;

            glLightf(i, GL_QUADRATIC_ATTENUATION, 0);
            glLightf(i, GL_CONSTANT_ATTENUATION, 1);
            glLightf(i, GL_LINEAR_ATTENUATION, (float)light.Attenuation);

            light.AbsoluteMatrix.Translation.CopyTo(_VectorArray, 1);
            glLightfv(i, GL_POSITION, _VectorArray);

            light.Color.CopyTo(_VectorArray);
            glLightfv(i, GL_AMBIENT, _VectorArray);
            glLightfv(i, GL_DIFFUSE, _VectorArray);
            glLightfv(i, GL_SPECULAR, _VectorArray);
            glLightfv(i, GL_EMISSION, _VectorArray);
        }

        private GeometryBuffer CreateUnitBounds()
        {
            double a = -0.5f;
            double b = -a;
            GeometryBuffer geometryBuffer = new GeometryBuffer(this);

            geometryBuffer.PrimitiveType = PrimitiveType.Lines;
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(a, a, a), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(a, b, a), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(b, b, a), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(b, a, a), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(a, a, b), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(a, b, b), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(b, b, b), new Vector3(), new Vector2()));
            geometryBuffer.Vertices.Add(new Vertex(new Vector3(b, a, b), new Vector3(), new Vector2()));
            geometryBuffer.Indices.AddRange(
                new int[] 
                { 
                    0, 1, 1, 2, 2, 3, 3, 0,
                    4, 5, 5, 6, 6, 7, 7, 4,
                    0, 4, 1, 5, 2, 6, 3, 7
                }
                );
            return geometryBuffer;
        }

        [DllImport("gdi32.dll")]
        private static extern bool SwapBuffers(IntPtr hDC);

        [DllImport("gdi32.dll")]
        private static extern int ChoosePixelFormat(IntPtr hDC, ref PIXELFORMATDESCRIPTOR pixelFormatDescriptor);

        [DllImport("gdi32.dll", EntryPoint = "SetPixelFormat")]
        private static extern bool _SetPixelFormat(IntPtr hDC, int pixelFormat, ref PIXELFORMATDESCRIPTOR pixelFormatDescriptor);

        private static bool SetPixelFormat(IntPtr deviceContext, int pixelFormat, ref PIXELFORMATDESCRIPTOR pixelFormatDescriptor)
        {
            LoadLibrary("opengl32.dll");
            return _SetPixelFormat(deviceContext, pixelFormat, ref pixelFormatDescriptor);
        }

        [DllImport("kernel32.dll")]
        private static extern IntPtr LoadLibrary(string fileName);

        [DllImport("user32.dll")]
        private static extern IntPtr GetDC(IntPtr hWND);

        [DllImport("user32.dll")]
        private static extern bool ReleaseDC(IntPtr hWND, IntPtr hDC);

        [DllImport("opengl32.dll")]
        private static extern IntPtr wglCreateContext(IntPtr hDC);

        [DllImport("opengl32.dll")]
        private static extern bool wglMakeCurrent(IntPtr hDC, IntPtr hGLRC);

        [DllImport("opengl32.dll")]
        private static extern bool wglDeleteContext(IntPtr hGLRC);

        private delegate bool wglSwapIntervalEXT(int interval);

        private static wglSwapIntervalEXT _wglSwapIntervalEXT = null;

        private static void LoadExtensions()
        {
            if (_wglSwapIntervalEXT == null)
            {
                IntPtr address = wglGetProcAddress("wglSwapIntervalEXT");

                if (address.ToInt32() != 0)
                    _wglSwapIntervalEXT = Marshal.GetDelegateForFunctionPointer(address, typeof(wglSwapIntervalEXT)) as wglSwapIntervalEXT;
            }
        }

        [DllImport("opengl32.dll")]
        private extern static IntPtr wglGetProcAddress(string s);

        [DllImport("opengl32.dll")]
        private extern static void glEnable(int cap);

        [DllImport("opengl32.dll")]
        private extern static void glDisable(int cap);

        [DllImport("opengl32.dll")]
        private extern static void glViewport(int x, int y, int width, int height);

        [DllImport("opengl32.dll")]
        private extern static void glDepthMask(int flag);

        [DllImport("opengl32.dll")]
        private extern static void glBlendFunc(int sourceFactor, int destFactor);

        [DllImport("opengl32.dll")]
        private extern static void glDepthFunc(int func);

        [DllImport("opengl32.dll")]
        private extern static void glFrontFace(int winding);

        [DllImport("opengl32.dll")]
        private extern static void glMatrixMode(int mode);

        [DllImport("opengl32.dll")]
        private extern static void glLoadIdentity();

        [DllImport("opengl32.dll")]
        private extern static void glMultMatrixd(double[] matrix);

        [DllImport("opengl32.dll")]
        private extern static void glPolygonMode(int face, int mode);

        [DllImport("opengl32.dll")]
        private extern static void glLightf(int light, int name, float value);

        [DllImport("opengl32.dll")]
        private extern static void glLightfv(int light, int name, float[] value);

        [DllImport("opengl32.dll")]
        private extern static void glColorMaterial(int face, int flag);

        [DllImport("opengl32.dll")]
        private extern static void glMaterialfv(int face, int name, float[] value);

        [DllImport("opengl32.dll")]
        private extern static void glMaterialf(int face, int name, float value);

        [DllImport("opengl32.dll")]
        private extern static void glLineWidth(float width);

        [DllImport("opengl32.dll")]
        private extern static void glPointSize(float size);

        [DllImport("opengl32.dll")]
        private extern static void glClearColor(float r, float g, float b, float a);

        [DllImport("opengl32.dll")]
        private extern static void glClear(int flags);

        [DllImport("opengl32.dll")]
        private extern static void glBegin(int primType);

        [DllImport("opengl32.dll")]
        private extern static void glColor4d(double r, double g, double b, double a);

        [DllImport("opengl32.dll")]
        private extern static void glNormal3d(double x, double y, double z);

        [DllImport("opengl32.dll")]
        private extern static void glTexCoord2d(double s, double t);

        [DllImport("opengl32.dll")]
        private extern static void glVertex3d(double x, double y, double z);

        [DllImport("opengl32.dll")]
        private extern static void glEnd();

        [DllImport("opengl32.dll")]
        private extern static void glPolygonOffset(float factor, float units);

        [DllImport("opengl32.dll")]
        internal extern static void glBindTexture(int mode, uint texture);

        [DllImport("opengl32.dll")]
        private extern static byte glIsTexture(uint texture);

        [DllImport("opengl32.dll")]
        private extern static void glGenTextures(int count, out uint texture);

        [DllImport("opengl32.dll")]
        private extern static void glDeleteTextures(int count, ref uint texture);

        [DllImport("opengl32.dll")]
        private extern static byte glIsList(uint list);

        [DllImport("opengl32.dll")]
        private extern static uint glGenLists(int range);

        [DllImport("opengl32.dll")]
        private extern static void glNewList(uint list, int mode);

        [DllImport("opengl32.dll")]
        private extern static void glEndList();

        [DllImport("opengl32.dll")]
        private extern static void glDeleteLists(uint list, int range);

        [DllImport("opengl32.dll")]
        private extern static void glCallList(uint list);

        [DllImport("opengl32.dll")]
        private extern static void glGetIntegerv(int param, int[] args);

        [DllImport("glu32.dll")]
        private extern static void gluBuild2DMipmaps(int target, int intFormat, int w, int h, int format, int type, byte[] data);
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct PIXELFORMATDESCRIPTOR
    {
        public Int16 nSize;
        public Int16 nVersion;
        public Int32 dwFlags;
        public Byte iPixelType;
        public Byte cColorBits;
        public Byte cRedBits;
        public Byte cRedShift;
        public Byte cGreenBits;
        public Byte cGreenShift;
        public Byte cBlueBits;
        public Byte cBlueShift;
        public Byte cAlphaBits;
        public Byte cAlphaShift;
        public Byte cAccumBits;
        public Byte cAccumRedBits;
        public Byte cAccumGreenBits;
        public Byte cAccumBlueBits;
        public Byte cAccumAlphaBits;
        public Byte cDepthBits;
        public Byte cStencilBits;
        public Byte cAuxBuffers;
        public Byte iLayerType;
        public Byte bReserved;
        public Int32 dwLayerMask;
        public Int32 dwVisibleMask;
        public Int32 dwDamageMask;
    }
}
