﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Lib3D
{
    public class SceneNode : IEnumerable
    {
        private GraphicsSystem _GraphicsSystem;
        private SceneNode _Parent;
        private LinkedList<SceneNode> _Children;
        private LinkedListNode<SceneNode> _Link;

        public SceneNode(GraphicsSystem graphicsSystem)
        {
            _GraphicsSystem = graphicsSystem;
            _Parent = null;
            _Children = new LinkedList<SceneNode>();
            _Link = null;

            Name = "";
            Visible = true;
            Enabled = true;
            DrawBounds = false;
            BoundsColor = new Vector4(1, 1, 1, 1);
            Tag = null;
            Materials = new List<Material>(new Material[] { new Material(_GraphicsSystem) });
            ModelMatrix = Matrix.CreateIdentity();
            Bounds = AABBox.CreateEmpty();
        }

        public SceneNode(SceneNode sceneNode)
        {
            _GraphicsSystem = sceneNode._GraphicsSystem;
            _Parent = null;
            _Children = new LinkedList<SceneNode>();
            _Link = null;

            Name = sceneNode.Name;
            Visible = sceneNode.Visible;
            Enabled = sceneNode.Enabled;
            DrawBounds = sceneNode.DrawBounds;
            BoundsColor = sceneNode.BoundsColor;
            Tag = sceneNode.Tag;
            ModelMatrix = sceneNode.ModelMatrix;
            Materials = new List<Material>(sceneNode.Materials.Count);
            foreach (Material material in sceneNode.Materials)
                Materials.Add(new Material(material));
            Bounds = sceneNode.Bounds;
            foreach (SceneNode childNode in sceneNode)
                AddChild(childNode.Clone());
        }

        public GraphicsSystem GraphicsSystem
        {
            get { return _GraphicsSystem; }
        }

        public string Name
        {
            get;
            set;
        }

        public bool Visible
        {
            get;
            set;
        }

        public bool Enabled
        {
            get;
            set;
        }

        public bool DrawBounds
        {
            get;
            set;
        }

        public Vector4 BoundsColor
        {
            get;
            set;
        }

        public object Tag
        {
            get;
            set;
        }

        public SceneNode Root
        {
            get
            {
                SceneNode root = this;

                if (Parent != null)
                    root = Parent.Root;
                return root;
            }
        }

        public SceneNode Parent
        {
            get { return _Parent; }
        }

        public SceneNode First
        {
            get
            {
                SceneNode first = null;

                if (_Children.First != null)
                    first = _Children.First.Value;
                return first;
            }
        }

        public SceneNode Last
        {
            get
            {
                SceneNode last = null;

                if (_Children.Last != null)
                    last = _Children.Last.Value;
                return last;
            }
        }

        public SceneNode Previous
        {
            get
            {
                SceneNode previous = null;

                if (_Link != null && _Link.Previous != null)
                    previous = _Link.Previous.Value;
                return previous;
            }
        }

        public SceneNode Next
        {
            get
            {
                SceneNode next = null;

                if (_Link != null && _Link.Next != null)
                    next = _Link.Next.Value;
                return next;
            }
        }

        public List<Material> Materials
        {
            get;
            set;
        }

        public Matrix ModelMatrix
        {
            get;
            set;
        }

        public Matrix AbsoluteMatrix
        {
            get
            {
                Matrix matrix = ModelMatrix;

                if (Parent != null)
                    matrix = Parent.AbsoluteMatrix * matrix;
                return matrix;
            }
        }

        public int Count
        {
            get
            {
                int count = 1;

                foreach (SceneNode sceneNode in this)
                    count += sceneNode.Count;
                return count;
            }
        }

        public AABBox Bounds
        {
            get;
            set;
        }

        public IEnumerator GetEnumerator()
        {
            return _Children.GetEnumerator();
        }

        public void AddChild(SceneNode sceneNode)
        {
            sceneNode.Detach();
            sceneNode._Parent = this;
            sceneNode._Link = _Children.AddLast(sceneNode);
        }

        public void Detach()
        {
            if (_Parent != null)
            {
                _Parent._Children.Remove(this);
                _Parent = null;
                _Link = null;
            }
        }

        public void ClearChildren()
        {
            while (_Children.Count != 0)
                _Children.First.Value.Detach();
        }

        public SceneNode Clone()
        {
            return Activator.CreateInstance(GetType(), this) as SceneNode;
        }

        public void SetLightingEnabled(bool enabled)
        {
            foreach (Material material in Materials)
                material.LightingEnabled = enabled;
            foreach (SceneNode sceneNode in this)
                sceneNode.SetLightingEnabled(enabled);
        }

        public void CalcBounds()
        {
            Matrix matrix = AbsoluteMatrix;

            Bounds = AABBox.CreateEmpty();
            foreach (Material material in Materials)
            {
                if (material.GeometryBuffer.Bounds.IsEmpty)
                    material.GeometryBuffer.CalcBounds();
                Bounds += matrix * material.GeometryBuffer.Bounds;
            }
            foreach (SceneNode sceneNode in this)
            {
                sceneNode.CalcBounds();
                Bounds += sceneNode.Bounds;
            }
        }

        protected virtual void OnUpdate()
        {
        }

        internal void Update()
        {
            OnUpdate();
            foreach (SceneNode sceneNode in this)
                sceneNode.Update();
        }
    }
}
