﻿using System;
using System.IO;
using System.Xml;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Lib3D;

namespace XMLScene
{
    public partial class Form1 : Form
    {
        GraphicsSystem _GraphicsSystem;
        Vector2 _LastMouse;
        MouseButtons _Button;
        TextureLoader _TexLoader;
        XmlReader _Reader;

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            _GraphicsSystem = new GraphicsSystem(_SceneView);
            _LastMouse = new Vector2(0, 0);
            _Button = MouseButtons.None;
            _TexLoader = new TextureLoader(_GraphicsSystem);
            _Timer.Start();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            DisposeGeometry(_GraphicsSystem.RootNode);
            _TexLoader.Dispose();
            _GraphicsSystem.Dispose();
        }

        private void _SceneView_MouseDown(object sender, MouseEventArgs e)
        {
            _Button = e.Button;
            _LastMouse = new Vector2(e.X, e.Y);
        }

        private void _SceneView_MouseUp(object sender, MouseEventArgs e)
        {
            _Button = MouseButtons.None;
        }

        private void _SceneView_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.None)
                return;

            Vector2 mouse = new Vector2(e.X, e.Y);
            Vector2 d = mouse - _LastMouse;

            _LastMouse = mouse;
            if (e.Button == MouseButtons.Middle)
                _GraphicsSystem.Camera.RotateAroundPosition(d);
            else
            {
                Vector3 f = Vector3.Normalize(_GraphicsSystem.Camera.Target - _GraphicsSystem.Camera.Position);
                Vector3 u = Vector3.Normalize(_GraphicsSystem.Camera.Up);
                Vector3 r = Vector3.Normalize(Vector3.Cross(f, u));

                d.Y = -d.Y;
                if (e.Button == MouseButtons.Left)
                    _GraphicsSystem.Camera.Position += f * d.Y + r * d.X;
                else
                    _GraphicsSystem.Camera.Position += u * d.Y;
                _GraphicsSystem.Camera.Target = _GraphicsSystem.Camera.Position + f;
                _GraphicsSystem.Camera.Up = Vector3.Normalize(Vector3.Cross(r, f));
            }
        }

        private void _FileOpenMenuItem_Click(object sender, EventArgs e)
        {
            if (_OpenFileDialog.ShowDialog() == DialogResult.OK)
                LoadScene(_OpenFileDialog.FileName);
        }

        private void _Timer_Tick(object sender, EventArgs e)
        {
            if (_SceneTree.SelectedNode == null)
                SetDrawBounds(_GraphicsSystem.RootNode, null);
            else
                SetDrawBounds(_GraphicsSystem.RootNode, _SceneTree.SelectedNode.Tag as SceneNode);
            _GraphicsSystem.Draw();
        }

        private void CompileGeometry(SceneNode sceneNode)
        {
            if (sceneNode != null)
            {
                foreach (Material material in sceneNode.Materials)
                    material.GeometryBuffer.Compile();
                foreach (SceneNode childNode in sceneNode)
                    CompileGeometry(childNode);
            }
        }

        private void DisposeGeometry(SceneNode sceneNode)
        {
            if (sceneNode != null)
            {
                foreach (Material material in sceneNode.Materials)
                    material.GeometryBuffer.Dispose();
                foreach (SceneNode childNode in sceneNode)
                    DisposeGeometry(childNode);
            }
        }

        private void SetDrawBounds(SceneNode sceneNode, SceneNode target)
        {
            if (sceneNode != null)
            {
                sceneNode.DrawBounds = target == sceneNode;
                foreach (SceneNode childNode in sceneNode)
                    SetDrawBounds(childNode, target);
            }
        }

        private void LoadScene(string file)
        {
            string caseFile = file;

            Text = "XML Scene";
            DisposeGeometry(_GraphicsSystem.RootNode);

            _SceneTree.Nodes.Clear();
            _GraphicsSystem.RootNode = new SceneNode(_GraphicsSystem);
            _TexLoader.Clear();

            try
            {
                file = file.ToLower();
                if (Path.GetExtension(file) == ".obj")
                    LoadOBJScene(file);
                else
                    LoadXMLScene(file);

                CompileGeometry(_GraphicsSystem.RootNode);
                BuildSceneTree(_GraphicsSystem.RootNode, _SceneTree.Nodes);
                Text = "XML Scene - " + caseFile;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                Console.WriteLine(ex.StackTrace);
                MessageBox.Show("Failed to load scene\r\n" + ex.Message);
            }
        }

        private void LoadOBJScene(string file)
        {
            OBJSceneNode objSceneNode = new OBJSceneNode(file, _TexLoader);

            objSceneNode.CalcBounds();

            AABBox bounds = objSceneNode.Bounds;
            Vector3 dim = bounds.Dimension;
            double max = Math.Max(Math.Max(dim.X, dim.Y), dim.Z);
            Vector3 scale = new Vector3(100, 100, 100) / max;
            Matrix scaleMatrix = Matrix.CreateScaling(scale);

            bounds = scaleMatrix * bounds;

            objSceneNode.Scale = scale;
            objSceneNode.Position = -bounds.Center;

            _GraphicsSystem.RootNode.AddChild(new LightSceneNode(_GraphicsSystem, new Vector3(-500, 500, -500), new Vector4(1, 1, 1, 1), 0.0001));
            _GraphicsSystem.RootNode.AddChild(new LightSceneNode(_GraphicsSystem, new Vector3(500, 500, 500), new Vector4(1, 1, 1, 1), 0.0001));
            _GraphicsSystem.RootNode.AddChild(objSceneNode);
            _GraphicsSystem.Camera.Position = new Vector3(100, 100, 100);
            _GraphicsSystem.Camera.Target = new Vector3(0, 0, 0);
            _GraphicsSystem.Camera.Up = new Vector3(-1, 1, -1);

            _GraphicsSystem.RootNode.First.Name = "LightSceneNode";
            _GraphicsSystem.RootNode.First.Next.Name = "LightSceneNode";
        }

        private void LoadXMLScene(string file)
        {
            XmlReaderSettings settings = new XmlReaderSettings();

            settings.IgnoreComments = true;
            settings.IgnoreWhitespace = true;
            settings.IgnoreProcessingInstructions = true;

            _Reader = null;
            try
            {
                _Reader = XmlReader.Create(file, settings);
                _Reader.ReadStartElement();

                ReadCamera();
                while (_Reader.IsStartElement())
                    _GraphicsSystem.RootNode.AddChild(ReadSceneNode());
                _Reader.ReadEndElement();
            }
            finally
            {
                if (_Reader != null)
                    _Reader.Close();
                _Reader = null;
            }
        }

        private void ReadCamera()
        {
            _Reader.ReadStartElement();

            _GraphicsSystem.Camera.Position = new Vector3(_Reader.ReadElementContentAsString());
            _GraphicsSystem.Camera.Target = new Vector3(_Reader.ReadElementContentAsString());
            _GraphicsSystem.Camera.Up = new Vector3(_Reader.ReadElementContentAsString());

            _Reader.ReadEndElement();
        }

        private SceneNode ReadSceneNode()
        {
            if (_Reader.Name == "LightSceneNode")
                return ReadLightSceneNode();
            else if (_Reader.Name == "TransformSceneNode")
                return ReadTransformSceneNode();
            else if (_Reader.Name == "OBJSceneNode")
                return ReadOBJSceneNode();
            else
                throw new Exception("Not a valid scene node element");
        }

        private LightSceneNode ReadLightSceneNode()
        {
            LightSceneNode lightSceneNode = new LightSceneNode(_GraphicsSystem);

            _Reader.ReadStartElement();

            lightSceneNode.Name = "LightSceneNode";
            lightSceneNode.Position = new Vector3(_Reader.ReadElementContentAsString());
            lightSceneNode.Color = new Vector4(_Reader.ReadElementContentAsString());
            lightSceneNode.Attenuation = _Reader.ReadElementContentAsDouble();

            while (_Reader.IsStartElement())
                lightSceneNode.AddChild(ReadSceneNode());
            _Reader.ReadEndElement();

            return lightSceneNode;
        }

        private TransformSceneNode ReadTransformSceneNode()
        {
            TransformSceneNode transformSceneNode = new TransformSceneNode(_GraphicsSystem);

            _Reader.ReadStartElement();

            transformSceneNode.Name = "TransformSceneNode";
            transformSceneNode.Position = new Vector3(_Reader.ReadElementContentAsString());
            transformSceneNode.SetRotation(new Vector3(_Reader.ReadElementContentAsString()));
            transformSceneNode.Scale = new Vector3(_Reader.ReadElementContentAsString());

            while (_Reader.IsStartElement())
                transformSceneNode.AddChild(ReadSceneNode());
            _Reader.ReadEndElement();

            return transformSceneNode;
        }

        private OBJSceneNode ReadOBJSceneNode()
        {
            OBJSceneNode objSceneNode = null;

            _Reader.ReadStartElement();

            objSceneNode = new OBJSceneNode(_Reader.ReadElementContentAsString(), _TexLoader);
            objSceneNode.Position = new Vector3(_Reader.ReadElementContentAsString());
            objSceneNode.SetRotation(new Vector3(_Reader.ReadElementContentAsString()));
            objSceneNode.Scale = new Vector3(_Reader.ReadElementContentAsString());

            while (_Reader.IsStartElement())
                objSceneNode.AddChild(ReadSceneNode());
            _Reader.ReadEndElement();

            return objSceneNode;
        }

        private void BuildSceneTree(SceneNode sceneNode, TreeNodeCollection treeNodes)
        {
            foreach (SceneNode childNode in sceneNode)
            {
                TreeNode treeNode = new TreeNode(childNode.Name);

                treeNode.Tag = childNode;
                childNode.Tag = treeNode;
                BuildSceneTree(childNode, treeNode.Nodes);
                treeNodes.Add(treeNode);
            }
        }
    }
}
